
import struct
from pyee import EventEmitter
from .aircraft_json_parser import AircraftJsonParser
from .parser import ProtocolParser


class ControlParser(EventEmitter):
    def __init__(self):
        super().__init__()
        self.aircraft_data, self.controls, self.address_lookup = AircraftJsonParser().get_aircraft_controls()
        self.emit_queue = []
        self.data_array =[0] * 65536
        self.parser = ProtocolParser(self.handle_data_change_for_address, self.handle_sync_complete)

    def handle_incoming_dcs_bios_message(self, message):
        for byte in bytearray(message):
            self.parser.process_byte(byte)

    def parse_address_data(self,address, data):
        if address not in self.address_lookup:
            return
        
        for control in self.address_lookup[address]:
            for output in control['outputs']:
                if output['address'] == address:
                    if output['type'] == 'string':
                        data_values = self.data_array[address:address + output['max_length']]

                        byte_data = bytes(data_values)
                        utf8_string = byte_data.decode('utf-8')
                        control_value = utf8_string
                    else:
                        control_value = (data & output['mask']) >> output['shift_by']
                    
                    value_has_changed = output.get('value') != control_value
                    is_already_added = any(item['control'] == control and item['output'] == output for item in self.emit_queue)

                    if value_has_changed and not is_already_added:
                        output['value'] = control_value
                        self.emit_queue.append({'output': output, 'control': control})

    def handle_data_change_for_address(self, address, data):
        bytes_data = struct.pack('<H', data)
        self.data_array[address] = bytes_data[0]
        self.data_array[address + 1] =  bytes_data[1]
        self.parse_address_data(address, data)

    def handle_sync_complete(self):
        for item in self.emit_queue:
            identifier = item['control']['identifier'] + item['output']['suffix']
            self.emit(identifier, item.output.value, item.control, item.output)
            self.emit(identifier + ':' + item.output.value, item.control, item.output)
        self.emit_queue = []





from pyee import EventEmitter
from .control_parser import ControlParser
from .udp_listener import UdpListener
from .event_constants import NETWORK_DATA_MESSAGE


class DcsBiosConnector(EventEmitter):

    def __init__(self):
        super().__init__()
        self.updListener = UdpListener()
        self.control_parser = ControlParser()
    
        self.updListener.on(NETWORK_DATA_MESSAGE, self.control_parser.handle_incoming_dcs_bios_message)


    def connect(self):
        self.updListener.start()
    
    def send(self):
        print("not implemented yet")



class ProtocolParser:
	def __init__(self, process_address_data_change_callback, sync_callback ):
		self.__state = "WAIT_FOR_SYNC"
		self.__sync_byte_count = 0
		self.__address = 0
		self.__count = 0
		self.__data = 0
		self.data_change_callback = process_address_data_change_callback
		self.sync_callback = sync_callback
		
	def process_byte(self, c):
		if self.__state == "ADDRESS_LOW":
			self.__address = c
			self.__state = "ADDRESS_HIGH"
		elif self.__state == "ADDRESS_HIGH":
			self.__address += c*256
			if self.__address != 0x5555:
				self.__state = "COUNT_LOW"
			else:
				self.__state = "WAIT_FOR_SYNC"
		elif self.__state == "COUNT_LOW":
			self.__count = c
			self.__state = "COUNT_HIGH"
		elif self.__state == "COUNT_HIGH":
			self.__count += 256*c
			self.__state = "DATA_LOW"
		elif self.__state == "DATA_LOW":
			self.__data = c
			self.__count -= 1
			self.__state = "DATA_HIGH"
		elif self.__state == "DATA_HIGH":
			self.__data += 256*c
			self.__count -= 1
			self.data_change_callback(self.__address, self.__data)
			self.__address += 2
			if self.__count == 0:
				self.__state = "ADDRESS_LOW"
			else:
				self.__state = "DATA_LOW"
			
			
		if c == 0x55:
			self.__sync_byte_count += 1
		else:
			self.__sync_byte_count = 0
		
		if self.__sync_byte_count == 4:
			self.__state = "ADDRESS_LOW"
			self.__sync_byte_count = 0
			self.sync_callback()



from pyee import EventEmitter

import socket
import struct
import threading
from .event_constants import NETWORK_DATA_MESSAGE

MULTICAST_GROUP = '239.255.50.10'
PORT = 5010
LOOPBACK_INTERFACE  = ""

class UdpListener(EventEmitter):
    def __init__(self):
        super().__init__()

    def udp_listener(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((LOOPBACK_INTERFACE, PORT))

        mreq = struct.pack('4sl', socket.inet_aton(MULTICAST_GROUP), socket.INADDR_ANY)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        print(f"Listening for data on {MULTICAST_GROUP}:{PORT}...")

        while True:
            data, addr = sock.recvfrom(65536)  # Buffer size of 65536 bytes
            self.emit(NETWORK_DATA_MESSAGE, data)

    def start(self):
        self.thread = threading.Thread(target=self.udp_listener)
        self.thread.daemon = True  # Makes sure the thread exits when the main program does
        self.thread.start()


from setuptools import setup, find_packages

setup(
    name='dcs_bios_connector',                  # Package name
    version='0.1.11',                    # Version
    description='Allows easy access to dcs bios state and sending of commands',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Papi Planes',
    author_email='papiplanesfly@gmail.com',
    url='https://github.com/papiplanes/dcs_bios_connector',
    packages=find_packages(), 
    package_data={
        'dcs_bios_connector': ['json/*.json'],
    },
    install_requires=[],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import os
import json

class AircraftJsonParser:
    def __init__(self):
        self.directory = './json'
        self.files = self.get_json_file_paths(self.directory)
    
    def get_aircraft_controls(self):
        return  JsonParser().parse_files(self.files)

    def get_json_file_paths(self, directory):
            """
            Get a list of relative paths to JSON files under the specified directory.

            :param directory: The directory to search for JSON files.
            :return: A list of relative paths to JSON files.
            """
            json_file_paths = []
            
            # Traverse the directory
            for root, dirs, files in os.walk(directory):
                for file in files:
                    if file.endswith('.json'):
                        # Get the absolute path to the JSON file
                        absolute_path = os.path.join(root, file)
                        json_file_paths.append(absolute_path)
            
            return json_file_paths


class JsonParser:
    @staticmethod
    def parse_files(files, aircraft_data=None, controls=None, address_lookup=None):
        if aircraft_data is None:
            aircraft_data = {}
        if controls is None:
            controls = {}
        if address_lookup is None:
            address_lookup = {}

        for file in files:
            # Get the aircraft name from the filename (without extension)
            aircraft_name = os.path.basename(file).replace('.json', '')
            
            # Load JSON data
            with open(file, 'r') as f:
                json_data = json.load(f)
            
            aircraft_data[aircraft_name] = json_data

            # Parse the aircraft's controls
            JsonParser._parse_json(json_data, controls, address_lookup)

        return aircraft_data, controls, address_lookup

    @staticmethod
    def _parse_json(json_data, controls, address_lookup):
        # Aircraft data is structured as Category -> [Array of Controls] -> [Array of Outputs]. We need to go 2 levels deep
        # to get the outputs.
        for category in JsonParser._get_values(json_data):
            for control in JsonParser._get_values(category):
                controls[control['identifier']] = control

                # Get the outputs for each control and add them to the address lookup
                for output in control['outputs']:
                    # Create the list at the control's address if it doesn't exist
                    if output['address'] not in address_lookup:
                        address_lookup[output['address']] = []
                    address_lookup[output['address']].append(control)

    @staticmethod
    def _get_values(obj):
        # In Python 3, we can use the .values() method directly
        if isinstance(obj, dict):
            return list(obj.values())
        else:
            return []